#include <cuda_runtime.h>
#include <stdio.h>

#define USEARRAY 0
#if USEARRAY
texture<ushort1, 2, cudaReadModeElementType> tex;
#else
texture<ushort1, 1, cudaReadModeElementType> tex;
#endif

const int sizeX = 960;
const int sizeY = 960;

__global__ void testkernel ( unsigned short* dst, int imageW, int imageH, int* visited )
{
	const int ix = blockDim.x * blockIdx.x + threadIdx.x;
	const int iy = blockDim.y * blockIdx.y + threadIdx.y;

	if ( ix < imageW && iy < imageH)// && ix > 1 && iy > 1  )
	{
		visited[imageW * iy + ix]++;
#if USEARRAY
		dst[imageW * iy + ix] = tex2D(tex, iy, ix).x;
#else
		dst[imageW * iy + ix] = tex1D(tex, imageW * iy + ix).x;
#endif
	}
}

int main ( int argc, char** argv )
{
	cudaError_t err;
	printf("Max linear mem tex size   : %i"
		   "\nWe have tex size in bytes : %i\n", 1<<27, sizeX * sizeY * sizeof(unsigned short));
	
	int requiredSpace = sizeX * sizeY * sizeof(ushort1);
	
	unsigned short* devMemSrc;
	unsigned short* devMemDst;
	unsigned short* hstMemDst;
	int* devMemVis;
	int* hstMemVis;
	
	cudaArray* cuArray;
	struct cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<ushort1>();

	cudaMalloc((void**)&devMemSrc, requiredSpace);
	cudaMalloc((void**)&devMemDst, requiredSpace);
	cudaMallocHost((void**)&hstMemDst, requiredSpace);
	cudaMalloc((void**)&devMemVis, sizeX * sizeY * sizeof(int));
	cudaMallocHost((void**)&hstMemVis, sizeX * sizeY * sizeof(int));
	cudaMallocArray(&cuArray, &channelDesc, 960, 960);
	

	cudaMemset((void*)devMemSrc, 1, requiredSpace);
	cudaMemset((void*)devMemVis, 0, sizeX * sizeY * sizeof(int));

#if USEARRAY
	cudaMemcpyToArray(cuArray, 0, 0, devMemSrc, requiredSpace, cudaMemcpyDeviceToDevice);
	cudaBindTextureToArray(tex, cuArray);
#else
	cudaBindTexture(0, tex, devMemSrc, requiredSpace);
#endif
	
	int blocksize = 8;
	int gridSizeX = 120;
	int gridSizeY = 120;
	dim3 thread(blocksize, blocksize);
	dim3 grid(gridSizeX, gridSizeY);
	
	testkernel<<<grid, thread>>>(devMemDst, sizeX, sizeY, devMemVis);
	if ( (err = cudaGetLastError()) != cudaSuccess )
		printf("<<< ERROR: %s\n", cudaGetErrorString(err));
	
	cudaMemcpy((void*)hstMemDst, (void*)devMemDst, requiredSpace, cudaMemcpyDeviceToHost);
	cudaMemcpy((void*)hstMemVis, (void*)devMemVis, sizeX * sizeY * sizeof(int), cudaMemcpyDeviceToHost);
	
	int counter = 0;
	for ( int y = 0; y < sizeY; y++)
		for ( int x = 0; x < sizeX; x++ )
			if ( hstMemVis[sizeX * y + x] > 0 )// && counter++ < 100 )
			{
				counter++;
				//printf("Visited x=%i y=%i : %i times\n", x, y, hstMemVis[sizeX * y + x]);
			}
	printf("#values visited: %i\n", counter);

	counter= 0;
	for ( int y = 0; y < sizeY; y++)
		for ( int x = 0; x < sizeX; x++ )
			if ( hstMemDst[sizeX * y + x] != 0 )
			{
				//printf("Value %i,%i: %i \n", x, y, hstMemDst[sizeX * y + x]);
				counter++;
			}
	
	printf("#values set: %i\n", counter);
	
}




